@@ -1,4 +1,4 @@
-{-# LANGUAGE ExplicitNamespaces #-}
+{-# LANGUAGE ConstraintKinds #-}
 {-# LANGUAGE GADTs #-}
 {-# LANGUAGE InstanceSigs #-}
 {-# LANGUAGE OverloadedStrings #-}
@@ -10,6 +10,11 @@
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE UndecidableInstances #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE DataKinds              #-}
+{-# LANGUAGE TypeFamilies           #-}
+{-# LANGUAGE PolyKinds #-}
+{-# LANGUAGE AllowAmbiguousTypes #-}
+{-# LANGUAGE LambdaCase #-}
 module DataFrame.Internal.Column where
 
 import qualified Data.ByteString.Char8 as C
@@ -24,19 +29,20 @@ import qualified Data.Vector.Unboxed as VU
 import qualified Data.Vector.Unboxed.Mutable as VUM
 
 import Control.Monad.ST (runST)
-import DataFrame.Internal.Function
 import DataFrame.Internal.Types
 import DataFrame.Internal.Parsing
 import Data.Int
 import Data.Maybe
+import Data.Proxy
 import Data.Text.Encoding (decodeUtf8Lenient)
 import Data.Type.Equality (type (:~:)(Refl), TestEquality (..))
-import Data.Typeable (Typeable)
+import Data.Typeable (Typeable, cast)
 import Data.Word
 import Type.Reflection
 import Unsafe.Coerce (unsafeCoerce)
 import DataFrame.Errors
 import Control.Exception (throw)
+import Data.Kind (Type, Constraint)
 
 -- | Our representation of a column is a GADT that can store data in either
 -- a vector with boxed elements or
@@ -50,6 +56,12 @@ data Column where
   MutableBoxedColumn :: Columnable a => VBM.IOVector a -> Column
   MutableUnboxedColumn :: (Columnable a, VU.Unbox a) => VUM.IOVector a -> Column
 
+data TypedColumn a where
+  TColumn :: Columnable a => Column -> TypedColumn a
+
+unwrapTypedColumn :: TypedColumn a -> Column
+unwrapTypedColumn (TColumn value) = value
+
 -- Functions about column metadata.
 isGrouped :: Column -> Bool
 isGrouped (GroupedBoxedColumn column) = True
@@ -74,6 +86,9 @@ columnTypeString column = case column of
   GroupedUnboxedColumn (column :: VB.Vector a) -> show (typeRep @a)
   GroupedOptionalColumn (column :: VB.Vector a) -> show (typeRep @a)
 
+instance (Show a) => Show (TypedColumn a) where
+  show (TColumn col) = show col
+
 instance Show Column where
   show :: Column -> String
   show (BoxedColumn column) = show column
@@ -112,47 +127,134 @@ instance Eq Column where
       Just Refl -> VB.map (L.sort . VG.toList) a == VB.map (L.sort . VG.toList) b
   (==) _ _ = False
 
-class (Columnable a) => Columnify a where
-  -- | Converts a boxed vector to a column making sure to put
-  -- the vector into an appropriate column type by reflection on the
-  -- vector's type parameter.
-  toColumn' :: VB.Vector a -> Column
+data Rep
+  = RBoxed
+  | RUnboxed
+  | ROptional
+  | RGBoxed
+  | RGUnboxed
+  | RGOptional
 
-instance (Columnable a) => Columnify (Maybe a) where
-  toColumn' = OptionalColumn
+type family If (cond :: Bool) (yes :: k) (no :: k) :: k where
+  If 'True  yes _  = yes
+  If 'False _   no = no
 
-instance (Columnable a) => Columnify (VB.Vector a) where
-  toColumn' = GroupedBoxedColumn
+type family Unboxable (a :: Type) :: Bool where
+  Unboxable Int    = 'True
+  Unboxable Int8   = 'True
+  Unboxable Int16  = 'True
+  Unboxable Int32  = 'True
+  Unboxable Int64  = 'True
+  Unboxable Word   = 'True
+  Unboxable Word8  = 'True
+  Unboxable Word16 = 'True
+  Unboxable Word32 = 'True
+  Unboxable Word64 = 'True
+  Unboxable Char   = 'True
+  Unboxable Bool   = 'True
+  Unboxable Double = 'True
+  Unboxable Float  = 'True
+  Unboxable _      = 'False
 
-instance (Columnable a, VU.Unbox a) => Columnify (VU.Vector a) where
-  toColumn' = GroupedUnboxedColumn
+-- | Compute the column representation tag for any ‘a’.
+type family KindOf a :: Rep where
+  KindOf (Maybe a)     = 'ROptional
+  KindOf (VB.Vector a) = 'RGBoxed
+  KindOf (VU.Vector a) = 'RGUnboxed
+  KindOf a             = If (Unboxable a) 'RUnboxed 'RBoxed
 
-instance {-# INCOHERENT #-} (Columnable a) => Columnify a where
-  toColumn' xs = case testEquality (typeRep @a) (typeRep @Int) of
-    Just Refl -> UnboxedColumn (VU.convert xs)
-    Nothing -> case testEquality (typeRep @a) (typeRep @Double) of
-      Just Refl -> UnboxedColumn (VU.convert xs)
-      Nothing -> case testEquality (typeRep @a) (typeRep @Float) of
-        Just Refl -> UnboxedColumn (VU.convert xs)
-        Nothing -> BoxedColumn xs
+class ColumnifyRep (r :: Rep) a where
+  toColumnRep :: VB.Vector a -> Column
 
-class (Columnable a) => ColumnifyList a where
-  -- | Converts a boxed vector to a column making sure to put
-  -- the vector into an appropriate column type by reflection on the
-  -- vector's type parameter.
-  toColumn :: [a] -> Column
+type Columnable a = (Columnable' a, ColumnifyRep (KindOf a) a, UnboxIf a, SBoolI (Unboxable a) )
 
-instance (Columnable a) => ColumnifyList (Maybe a) where
-  toColumn = OptionalColumn . VB.fromList
+instance (Columnable a, VU.Unbox a)
+      => ColumnifyRep 'RUnboxed a where
+  toColumnRep = UnboxedColumn . VU.convert
 
-instance {-# INCOHERENT #-} (Columnable a) => ColumnifyList a where
-  toColumn xs = case testEquality (typeRep @a) (typeRep @Int) of
-    Just Refl -> UnboxedColumn (VU.fromList xs)
-    Nothing -> case testEquality (typeRep @a) (typeRep @Double) of
-      Just Refl -> UnboxedColumn (VU.fromList xs)
-      Nothing -> case testEquality (typeRep @a) (typeRep @Float) of
-        Just Refl -> UnboxedColumn (VU.fromList xs)
-        Nothing -> BoxedColumn (VB.fromList xs)
+instance Columnable a
+      => ColumnifyRep 'RBoxed a where
+  toColumnRep = BoxedColumn
+
+instance Columnable a
+      => ColumnifyRep 'ROptional (Maybe a) where
+  toColumnRep = OptionalColumn
+
+instance Columnable a
+      => ColumnifyRep 'RGBoxed (VB.Vector a) where
+  toColumnRep = GroupedBoxedColumn
+
+instance (Columnable a, VU.Unbox a)
+      => ColumnifyRep 'RGUnboxed (VU.Vector a) where
+  toColumnRep = GroupedUnboxedColumn
+
+toColumn' ::
+  forall a. (Columnable a, ColumnifyRep (KindOf a) a)
+  => VB.Vector a -> Column
+toColumn' = toColumnRep @(KindOf a)
+
+toColumn ::
+  forall a. (Columnable a, ColumnifyRep (KindOf a) a)
+  => [a] -> Column
+toColumn = toColumnRep @(KindOf a) . VB.fromList
+
+data SBool (b :: Bool) where
+  STrue  :: SBool 'True
+  SFalse :: SBool 'False
+
+class SBoolI (b :: Bool) where
+  sbool :: SBool b          -- the run-time witness
+
+instance SBoolI 'True  where sbool = STrue
+instance SBoolI 'False where sbool = SFalse
+
+sUnbox :: forall a. SBoolI (Unboxable a) => SBool (Unboxable a)
+sUnbox = sbool @(Unboxable a)
+
+type family When (flag :: Bool) (c :: Constraint) :: Constraint where
+  When 'True  c = c
+  When 'False c = ()          -- empty constraint
+
+type UnboxIf a = When (Unboxable a) (VU.Unbox a)
+
+-- | Generic column transformation (no index).
+transform
+  :: forall b c.
+     ( Columnable b
+     , Columnable c
+     , UnboxIf c
+     , Typeable b
+     , Typeable c )
+  => (b -> c)
+  -> Column
+  -> Maybe Column
+transform f = \case
+  BoxedColumn (col :: VB.Vector a)
+    | Just Refl <- testEquality (typeRep @a) (typeRep @b)
+    -> Just (toColumn' @c (VB.map f col))
+    | otherwise -> Nothing
+  OptionalColumn (col :: VB.Vector a)
+    | Just Refl <- testEquality (typeRep @a) (typeRep @b)
+    -> Just (toColumn' @c (VB.map f col))
+    | otherwise -> Nothing
+  UnboxedColumn (col :: VU.Vector a)
+    | Just Refl <- testEquality (typeRep @a) (typeRep @b)
+    -> Just $ case sUnbox @c of
+                STrue  -> UnboxedColumn (VU.map f col)
+                SFalse -> toColumn' @c (VB.map f (VB.convert col))
+    | otherwise -> Nothing
+  GroupedBoxedColumn (col :: VB.Vector (VB.Vector a))
+    | Just Refl <- testEquality (typeRep @(VB.Vector a)) (typeRep @b)
+    -> Just (toColumn' @c (VB.map f col))
+    | otherwise -> Nothing
+  GroupedUnboxedColumn (col :: VB.Vector (VU.Vector a))
+    | Just Refl <- testEquality (typeRep @(VU.Vector a)) (typeRep @b)
+    -> Just (toColumn' @c (VB.map f col))
+    | otherwise -> Nothing
+  GroupedOptionalColumn (col :: VB.Vector (VB.Vector a))
+    | Just Refl <- testEquality (typeRep @(VB.Vector a)) (typeRep @b)
+    -> Just (toColumn' @c (VB.map f col))
+    | otherwise -> Nothing
 
 -- | Converts a an unboxed vector to a column making sure to put
 -- the vector into an appropriate column type by reflection on the
@@ -255,79 +357,34 @@ sortedIndexes asc (GroupedOptionalColumn column) = runST $ do
   return $ VU.generate (VG.length column) (\i -> fst (sorted VG.! i))
 {-# INLINE sortedIndexes #-}
 
--- Operations on a column that may change its type.
-
-instance Transformable Column where
-  transform :: forall b c . (Columnable b, Columnable c) => (b -> c) -> Column -> Maybe Column
-  transform f (BoxedColumn (column :: VB.Vector a)) = do
-    Refl <- testEquality (typeRep @a) (typeRep @b)
-    return (toColumn' (VB.map f column))
-  transform f (OptionalColumn (column :: VB.Vector a)) = do
-    Refl <- testEquality (typeRep @a) (typeRep @b)
-    return (toColumn' (VB.map f column))
-  transform f (UnboxedColumn (column :: VU.Vector a)) = do
-    Refl <- testEquality (typeRep @a) (typeRep @b)
-    return $ if testUnboxable (typeRep @c) then transformUnboxed f column else toColumn' (VB.map f (VB.convert column))
-  transform f (GroupedBoxedColumn (column :: VB.Vector (VB.Vector a))) = do
-    Refl <- testEquality (typeRep @(VB.Vector a)) (typeRep @b)
-    return (toColumn' (VB.map f column))
-  transform f (GroupedUnboxedColumn (column :: VB.Vector (VU.Vector a))) = do
-    Refl <- testEquality (typeRep @(VU.Vector a)) (typeRep @b)
-    return (toColumn' (VB.map f column))
-  transform f (GroupedOptionalColumn (column :: VB.Vector (VB.Vector a))) = do
-    Refl <- testEquality (typeRep @(VB.Vector a)) (typeRep @b)
-    return (toColumn' (VB.map f column))
-
 -- | Applies a function that returns an unboxed result to an unboxed vector, storing the result in a column.
-transformUnboxed :: forall a b . (Columnable a, VU.Unbox a, Columnable b) => (a -> b) -> VU.Vector a -> Column
-transformUnboxed f = itransformUnboxed (const f)
-
--- TODO: Make a type class with incoherent instances.
-itransformUnboxed :: forall a b . (Columnable a, VU.Unbox a, Columnable b) => (Int -> a -> b) -> VU.Vector a -> Column
-itransformUnboxed f column = case testEquality (typeRep @b) (typeRep @Int) of
-  Just Refl -> UnboxedColumn $ VU.imap f column
-  Nothing -> case testEquality (typeRep @b) (typeRep @Int8) of
-    Just Refl -> UnboxedColumn $ VU.imap f column
-    Nothing -> case testEquality (typeRep @b) (typeRep @Int16) of
-      Just Refl -> UnboxedColumn $ VU.imap f column
-      Nothing -> case testEquality (typeRep @b) (typeRep @Int32) of
-        Just Refl -> UnboxedColumn $ VU.imap f column
-        Nothing -> case testEquality (typeRep @b) (typeRep @Int64) of
-          Just Refl -> UnboxedColumn $ VU.imap f column
-          Nothing -> case testEquality (typeRep @b) (typeRep @Word8) of
-            Just Refl -> UnboxedColumn $ VU.imap f column
-            Nothing-> case testEquality (typeRep @b) (typeRep @Word16) of
-              Just Refl -> UnboxedColumn $ VU.imap f column
-              Nothing -> case testEquality (typeRep @b) (typeRep @Word32) of
-                Just Refl -> UnboxedColumn $ VU.imap f column
-                Nothing -> case testEquality (typeRep @b) (typeRep @Word64) of
-                  Just Refl -> UnboxedColumn $ VU.imap f column
-                  Nothing -> case testEquality (typeRep @b) (typeRep @Char) of
-                    Just Refl -> UnboxedColumn $ VU.imap f column
-                    Nothing -> case testEquality (typeRep @b) (typeRep @Bool) of
-                      Just Refl -> UnboxedColumn $ VU.imap f column
-                      Nothing -> case testEquality (typeRep @b) (typeRep @Float) of
-                        Just Refl -> UnboxedColumn $ VU.imap f column
-                        Nothing -> case testEquality (typeRep @b) (typeRep @Double) of
-                          Just Refl -> UnboxedColumn $ VU.imap f column
-                          Nothing -> case testEquality (typeRep @b) (typeRep @Word) of
-                            Just Refl -> UnboxedColumn $ VU.imap f column
-                            Nothing -> error "Result type is unboxed" -- since we only call this after confirming 
-
--- | tranform with index.
-itransform :: forall b c. (Columnable b, Columnable c) => (Int -> b -> c) -> Column -> Maybe Column
-itransform f (BoxedColumn (column :: VB.Vector a)) = do
-  Refl <- testEquality (typeRep @a) (typeRep @b)
-  return (toColumn' (VB.imap f column))
-itransform f (UnboxedColumn (column :: VU.Vector a)) = do
-  Refl <- testEquality (typeRep @a) (typeRep @b)
-  return $ if testUnboxable (typeRep @c) then itransformUnboxed f column else toColumn' (VB.imap f (VB.convert column))
-itransform f (GroupedBoxedColumn (column :: VB.Vector (VB.Vector a))) = do
-  Refl <- testEquality (typeRep @(VB.Vector a)) (typeRep @b)
-  return (toColumn' (VB.imap f column))
-itransform f (GroupedUnboxedColumn (column :: VB.Vector (VU.Vector a))) = do
-  Refl <- testEquality (typeRep @(VU.Vector a)) (typeRep @b)
-  return (toColumn' (VB.imap f column))
+itransform
+  :: forall b c. (Typeable b, Typeable c, Columnable b, Columnable c)
+  => (Int -> b -> c) -> Column -> Maybe Column
+itransform f = \case
+  BoxedColumn (col :: VB.Vector a)
+    | Just Refl <- testEquality (typeRep @a) (typeRep @b)
+    -> Just (toColumn' @c (VB.imap f col))
+    | otherwise -> Nothing
+  UnboxedColumn (col :: VU.Vector a)
+    | Just Refl <- testEquality (typeRep @a) (typeRep @b)
+    -> Just $
+        case sUnbox @c of
+          STrue  -> UnboxedColumn (VU.imap f col)
+          SFalse -> toColumn' @c (VB.imap f (VB.convert col))
+    | otherwise -> Nothing
+  OptionalColumn (col :: VB.Vector a)
+    | Just Refl <- testEquality (typeRep @a) (typeRep @b)
+    -> Just (toColumn' @c (VB.imap f col))
+    | otherwise -> Nothing
+  GroupedBoxedColumn (col :: VB.Vector a)
+    | Just Refl <- testEquality (typeRep @a) (typeRep @b)
+    -> Just (toColumn' @c (VB.imap f col))
+    | otherwise -> Nothing
+  GroupedUnboxedColumn (col :: VB.Vector a)
+    | Just Refl <- testEquality (typeRep @a) (typeRep @b)
+    -> Just (toColumn' @c (VB.imap f col))
+    | otherwise -> Nothing
 
 -- | Filter column with index.
 ifilterColumn :: forall a . (Columnable a) => (Int -> a -> Bool) -> Column -> Maybe Column
@@ -344,23 +401,6 @@ ifilterColumn f c@(GroupedUnboxedColumn (column :: VB.Vector b)) = do
   Refl <- testEquality (typeRep @a) (typeRep @b)
   return $ GroupedUnboxedColumn $ VG.ifilter f column
 
--- TODO: Expand this to use more predicates.
-ifilterColumnF :: Function -> Column -> Maybe Column
-ifilterColumnF (ICond (f :: Int -> a -> Bool)) c@(BoxedColumn (column :: VB.Vector b)) = do
-  Refl <- testEquality (typeRep @a) (typeRep @b)
-  return $ BoxedColumn $ VG.ifilter f column
-ifilterColumnF (ICond (f :: Int -> a -> Bool)) c@(UnboxedColumn (column :: VU.Vector b)) = do
-  Refl <- testEquality (typeRep @a) (typeRep @b)
-  return $ UnboxedColumn $ VG.ifilter f column
-ifilterColumnF (ICond (f :: Int -> a -> Bool)) c@(OptionalColumn (column :: VB.Vector b)) = do
-  Refl <- testEquality (typeRep @a) (typeRep @b)
-  return $ OptionalColumn $ VG.ifilter f column
-ifilterColumnF (ICond (f :: Int -> a -> Bool)) c@(GroupedBoxedColumn (column :: VB.Vector b)) = do
-  Refl <- testEquality (typeRep @a) (typeRep @b)
-  return $ GroupedBoxedColumn $ VG.ifilter f column
-ifilterColumnF (ICond (f :: Int -> a -> Bool)) c@(GroupedUnboxedColumn (column :: VB.Vector b)) = do
-  Refl <- testEquality (typeRep @a) (typeRep @b)
-  return $ GroupedUnboxedColumn $ VG.ifilter f column
 
 ifoldrColumn :: forall a b. (Columnable a, Columnable b) => (Int -> a -> b -> b) -> b -> Column -> Maybe b
 ifoldrColumn f acc c@(BoxedColumn (column :: VB.Vector d)) = do
@@ -430,6 +470,18 @@ zipColumns (UnboxedColumn column) (BoxedColumn other) = BoxedColumn (VB.generate
 zipColumns (UnboxedColumn column) (UnboxedColumn other) = UnboxedColumn (VG.zip column other)
 {-# INLINE zipColumns #-}
 
+zipWithColumns :: forall a b c . (Columnable a, Columnable b, Columnable c) => (a -> b -> c) -> Column -> Column -> Column
+zipWithColumns f (UnboxedColumn (column :: VU.Vector d)) (UnboxedColumn (other :: VU.Vector e)) = case testEquality (typeRep @a) (typeRep @d) of
+  Just Refl -> case testEquality (typeRep @b) (typeRep @e) of
+    Just Refl -> toColumn' $ VB.zipWith f (VG.convert column) (VG.convert other)
+    Nothing -> throw $ TypeMismatchException' (typeRep @b) (show $ typeRep @e) "" "zipWithColumns"
+  Nothing -> throw $ TypeMismatchException' (typeRep @a) (show $ typeRep @d) "" "zipWithColumns"
+zipWithColumns f left right = let
+    left' = toVector @a left
+    right' = toVector @b right
+  in toColumn' $ VB.zipWith f left' right' 
+{-# INLINE zipWithColumns #-}
+
 -- Functions for mutable columns (intended for IO).
 -- Clean this up.
 writeColumn :: Int -> T.Text -> Column -> IO (Either T.Text Bool)
