@@ -21,8 +21,7 @@ import DataFrame.Errors (DataFrameException(..))
 import DataFrame.Internal.Column
 import DataFrame.Internal.DataFrame (DataFrame(..), getColumn, empty)
 import DataFrame.Internal.Expression
-import DataFrame.Internal.Row (mkRowFromArgs)
-import DataFrame.Internal.Types (RowValue, toRowValue)
+import DataFrame.Internal.Row (mkRowFromArgs, RowValue, toRowValue)
 import DataFrame.Operations.Core
 import DataFrame.Operations.Transformations (apply)
 import Data.Function ((&))
